/* Process model C form file: HostRsvpDeamon.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "HostRsvpDeamon.pr.h"
FSM_EXT_DECS



/* Header block */

#include<stdio.h>
#include<errno.h>

#define ARRIVAL op_intrpt_type()==OPC_INTRPT_STRM
#define END_SIM     op_intrpt_type () ==  OPC_INTRPT_ENDSIM

#define Path_Msg_Refresh_timeout 4000

#define IPSTREAM 				0
#define APPL_STREAM				1

#define InLIST_SIZE     		10
#define TCP_PROTOCOL_ID			12
#define UDP_PROTOCOL_ID			20
#define RSVP_PROTOCOL_ID 		46
#define REFRESH_PERIOD     		300

#define RSVP_CURRENT_VER		0x01

#define PATH_MESSAGE			01
#define RESV_MESSAGE			02
#define PATH_ERROR				03
#define RESV_ERROR				04
#define PATH_TEAR				05
#define RESV_TEAR				06
#define RESV_CONF				07
#define SESSION_MSG				100
#define SENDER_MSG				101 
#define RESERVE_MSG				102
#define RELEASE_MSG				103

#define MAXNOOFSESSIONS			100

#define ON						1
#define OFF						0
#define VL						if(ON)
#define NOV						if(OFF)

#define UNDEFINED       -1
#define DEFAULT_RESERVE_TTL      7
#define MCAST_ADDRESS   500

#ifndef _DEBUG_FUNCS_H
#define _DEBUG_FUNCS_H
 
 
#define PrintErr(p)  \
            printf p ; printf(" : FileName = %s : LineNo = %d :\n",  __FILE__, __LINE__);  printf("ErrNo = %d
: ErrStr = %s :\n\n",  errno, strerror(errno));

#define MALLOC(sz)      calloc(1,sz)
#define Open(x,y,z) \
        open(x,y,z); \
        if( errno ) \
        { PrintErr(("Cannot open file %s ",x)); }
 
#define OpenE(x,y,z) \
        open(x,y,z); \
        if( errno ) \
        { PrintErr( ("Cannot open file %s ",x) ); exit(errno); }
 
#define Malloc(x) \
        malloc(x); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); }
 
#define MallocE(x) \
        malloc(x); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); exit(errno); }
 
#define Calloc(x,y) \
        calloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); }
 
#define CallocE(x,y) \
        calloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); exit(errno); }
 
#define Realloc(x,y) \
        realloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); }
#define ReallocE(x,y) \
        realloc(x,y); \
        if( errno ) \
        { PrintErr(("Cannot Allocate memory")); exit(errno); }
 
#define Read(x,y,z) \
        read(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); }
 
#define ReadE(x,y,z) \
        read(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); exit(errno); }
 
#define Write(x,y,z) \
        write(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); }
 
#define WriteE(x,y,z) \
        write(x,y,z); \
        if( errno ) \
        { PrintErr(("Read failed")); exit(errno); }
 
#define Fopen(x,y) \
        fopen(x,y); \
        if( errno ) \
        { PrintErr(("Cannot open file %s", x)); }
 
#define FopenE(x,y) \
        fopen(x,y); \
        if( errno ) \
        { PrintErr(("Cannot open file %s", x)); exit(errno); }
 
#define Fread(x,y,z,s) \
        fread(x,y,z,s); \
        if( errno ) \
        { PrintErr(("Cannot read file")); }
 
#define FreadE(x,y,z,s) \
        fread(x,y,z,s); \
        if( errno ) \
        { PrintErr(("Cannot read file")); exit(errno); }
 
#define Fwrite(x,y,z,s) \
        fwrite(x,y,z,s); \
        if( errno ) \
		 { PrintErr(("Cannot read file")); }
 
#define FwriteE(x,y,z,s) \
        fwrite(x,y,z,s); \
        if( errno ) \
        { PrintErr(("Cannot read file")); exit(errno); }
 
#define Memcpy(x,y,z) op_prg_mem_copy( y, x, z );
 
#endif

/*
** IP Address structure
*/
typedef struct {
	int NetNum;
	int NodeNum;
} IPAddress;

struct GenList {
	struct GenList *next, *curr;
	char *ele;
};

typedef	struct GenList GenList;

/* 
** Session Information
*/
typedef struct {
	IPAddress DestAddress;
	int ProtocolId;
	int DestPort;
} SessInfo;

/*
** Common Header
*/
typedef struct {
	int Version;
	int Msg_Type;
	int Send_TTL;
	double RSVP_Checksum;
	double RSVP_Length;
} CommonHdr;

/*
** Session object 
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress DestAddress;
	int Protocol_Id;
	int Flags;
	int DestPort;
} SessionObj;

/*
** RSVP Hop Object
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress Next_Hop_Address;
	IPAddress Logical_Interface_Handle;
} RsvpHopObj;

/*
** Time values object
*/
typedef struct {
	int Class_Num; 
	double Length;
	double Refresh_Period;
} TimeValObj;

/*
** Sender template
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress SrcAddress;
	int SrcPort;
} SenderTemplateObj;

/*
** Sender T-Spec object
*/
typedef struct {
	int Class_Num; 
	double Length;
	double req_bandwidth;
	double link_delay;
} SenderTspecObj;

/*
** FlowSpec Object
*/
typedef SenderTspecObj FlowSpecObj;

/*
** Filter Spec
*/
typedef SenderTemplateObj FilterSpecObj;

/*
** Style 
*/
typedef struct {
	int Class_Num;
	double Length;
	int Resv_Style;
} StyleObj;

/*
** Scope object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress SrcAddress;
} ScopeObj;

/*
** Reservation Confirm object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress Receiver_Address;
} ResvConfirmObj;

/*
** Error Spec object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress Error_Node_Address;
	int Flags;
	int Error_code;
} ErrorSpecObj;



/* 
** Sender Descriptor structure 
*/
typedef struct {

	/* SENDER_TEMPLATE structure */
  SenderTemplateObj Sender_Template_Object; 

	/* SENDER_TSPEC structure */
  SenderTspecObj Sender_Tspec_Object;     

}SenderDescriptor;

/* 
** Flow descriptor list structure 
*/
typedef struct {

	/* FLOWSPEC structure */
  FlowSpecObj Flowspec_Object;         

  /* this was added when it was known that FF style needs to be considered */
	/* FILTER_SPEC structure */
  FilterSpecObj Filter_Spec_Object;      

}FlowDescriptorList;

/* 
** PATH message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header;

	/* SESSION object structure */
  SessionObj Session_Object;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object;

	/* TIME_VALUES Object structure */
  TimeValObj Time_Values_Object;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor;

}PathMessage;

/* 
** RESV message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* TIME_VALUES Object structure */
  TimeValObj Time_Values_Object       ;

	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list[InLIST_SIZE];

}ResvMessage;

/* 
** PathTear message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor        ;

}PathTearMessage;

/* 
** ResvTear message structure 
*/
typedef struct {
	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list  ;

}ResvTearMessage;

/* 
** PathErr message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor        ;

}PathErrMessage;

/* 
** ResvErr message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object;

	/* STYLE Object structure */
  StyleObj Style_Object;

	/* Flow descriptor list structure */
  /*  ErrorFlowDescriptor error_flow_descriptor;  */

} ResvErrMessage;

/* 
** ResvConf message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;


	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list  ;

}ResvConfMessage;
	
typedef struct {
	int isfull;
	IPAddress DestAddress;
	int ProtocolId;
	int DestPort;
	void (*UpcallProc)();
} RsvpSession;

typedef struct {
	int SessionID;
	IPAddress SrcAddress;
	int SrcPort;
	SenderTemplateObj Sender_Template_Object;
	SenderTspecObj Sender_Tspec_Object;
	int TTL;
} RsvpSender;

typedef struct{
	int SessionID;
	FlowDescriptorList flow_descriptor_list[InLIST_SIZE];
} RsvpReserve;

/**
*** Function Prototypes
**/

int Session( IPAddress da, int pid, int dp, void (*Up)() );
void Sender( int sid, IPAddress sa, int sp, SenderTemplateObj sto,
			 SenderTspecObj sts, int ttl );
void Reserve( int sid, FlowDescriptorList * fdl );
void Release( int sid );
void HostInitialize( void );
void SendPathMsg( PathMessage pmEle, IPAddress sa );
void MakeSessionCall( Packet* Pkptr );
void MakeSenderCall( Packet* Pkptr );
void MakeReserveCall( Packet* Pkptr );



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	int                     		sv_SessionCallFlag;
	int                     		sv_SenderCallFlag;
	int                     		sv_ReleaseCallFlag;
	int                     		sv_ReserveCallFlag;
	int                     		sv_RecDataPacket;
	int                     		sv_RsvpUpCalls;
	int                     		sv_SessionId;
	int                     		sv_GrpId[InLIST_SIZE];
	int                     		sv_GrpNo;
	int                     		sv_LinkNo;
	Objid                   		sv_my_id;
	Objid                   		sv_ip_objid;
	int                     		sv_InStrm;
	int                     		sv_data_length;
	Packet*                 		sv_RecvPkptr;
	int                     		sv_NoOfLinks;
	IPAddress               		sv_HostIPAddress;
	IPAddress               		sv_LocalAddress[InLIST_SIZE];
	RsvpSession             		sv_rsvpSession[MAXNOOFSESSIONS];
	RsvpSender              		sv_rsvpSender[MAXNOOFSESSIONS];
	RsvpReserve             		sv_rsvpReserve[MAXNOOFSESSIONS];
	struct GenList          		sv_PSBList;
	struct GenList          		sv_RSBList;
	struct GenList          		sv_TCSBList;
	} HostRsvpDeamon_state;

#define pr_state_ptr            		((HostRsvpDeamon_state*) SimI_Mod_State_Ptr)
#define SessionCallFlag         		pr_state_ptr->sv_SessionCallFlag
#define SenderCallFlag          		pr_state_ptr->sv_SenderCallFlag
#define ReleaseCallFlag         		pr_state_ptr->sv_ReleaseCallFlag
#define ReserveCallFlag         		pr_state_ptr->sv_ReserveCallFlag
#define RecDataPacket           		pr_state_ptr->sv_RecDataPacket
#define RsvpUpCalls             		pr_state_ptr->sv_RsvpUpCalls
#define SessionId               		pr_state_ptr->sv_SessionId
#define GrpId                   		pr_state_ptr->sv_GrpId
#define GrpNo                   		pr_state_ptr->sv_GrpNo
#define LinkNo                  		pr_state_ptr->sv_LinkNo
#define my_id                   		pr_state_ptr->sv_my_id
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define InStrm                  		pr_state_ptr->sv_InStrm
#define data_length             		pr_state_ptr->sv_data_length
#define RecvPkptr               		pr_state_ptr->sv_RecvPkptr
#define NoOfLinks               		pr_state_ptr->sv_NoOfLinks
#define HostIPAddress           		pr_state_ptr->sv_HostIPAddress
#define LocalAddress            		pr_state_ptr->sv_LocalAddress
#define rsvpSession             		pr_state_ptr->sv_rsvpSession
#define rsvpSender              		pr_state_ptr->sv_rsvpSender
#define rsvpReserve             		pr_state_ptr->sv_rsvpReserve
#define PSBList                 		pr_state_ptr->sv_PSBList
#define RSBList                 		pr_state_ptr->sv_RSBList
#define TCSBList                		pr_state_ptr->sv_TCSBList




/* Process model interrupt handling procedure */



void
HostRsvpDeamon ()
	{
	Packet* ReceivedPkptr;
	Packet* UpcallPkptr;
	Packet* ApplPkptr;


	FSM_ENTER (HostRsvpDeamon)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (Init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "Init")
			{
			HostInitialize();
			SessionCallFlag = 0;
			SenderCallFlag = 0;
			ReserveCallFlag = 0;
			ReleaseCallFlag = 0;
			}


		/** state (Init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "Init")
			{
			}


		/** state (Init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,HostRsvpDeamon)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (ARRIVAL)
		FSM_TEST_COND (END_SIM)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 2, state2_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 7, state7_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (Arr) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "Arr")
			{
			InStrm = op_intrpt_strm();
			
			if( InStrm == APPL_STREAM )
			{
				int RecvCallType;
				RecvPkptr = op_pk_get( InStrm );
				op_pk_nfd_get(RecvPkptr, "ApplCallType", &RecvCallType);
				switch( RecvCallType )
				{
					case SESSION_MSG :
						VL	NOV	printf("Received Session call from Appl. \n");
						op_pk_nfd_get( RecvPkptr, "AppCall", &ApplPkptr );
						MakeSessionCall( ApplPkptr );
					break;
					case SENDER_MSG  :
						VL	NOV	printf("Received Sender call from Appl. \n");
						op_pk_nfd_get( RecvPkptr, "AppCall", &ApplPkptr );
						MakeSenderCall( ApplPkptr );
					break;
					case RESERVE_MSG :
						VL	NOV	printf("Received Reserve call from Appl. \n");
						op_pk_nfd_get( RecvPkptr, "AppCall", &ApplPkptr );
						MakeReserveCall( ApplPkptr ); 
					break;
					case RELEASE_MSG :
						op_pk_nfd_get( RecvPkptr, "AppCall", &ApplPkptr );
						/*  MakeReleaseCall( ApplPkptr );  */
					break;
				}
			}
			}


		/** state (Arr) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "Arr")
			{
			}


		/** state (Arr) transition processing **/
		FSM_INIT_COND (SessionCallFlag)
		FSM_TEST_COND (SenderCallFlag)
		FSM_TEST_COND (ReserveCallFlag)
		FSM_TEST_COND (ReleaseCallFlag)
		FSM_DFLT_COND
		FSM_TEST_LOGIC ("Arr")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 3, state3_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 4, state4_enter_exec, ;)
			FSM_CASE_TRANSIT (2, 5, state5_enter_exec, ;)
			FSM_CASE_TRANSIT (3, 6, state6_enter_exec, ;)
			FSM_CASE_TRANSIT (4, 1, state1_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (Session) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "Session")
			{
			}


		/** state (Session) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "Session")
			{
			}


		/** state (Session) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Sender) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "Sender")
			{
			}


		/** state (Sender) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "Sender")
			{
			}


		/** state (Sender) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Reserve) enter executives **/
		FSM_STATE_ENTER_FORCED (5, state5_enter_exec, "Reserve")
			{
			}


		/** state (Reserve) exit executives **/
		FSM_STATE_EXIT_FORCED (5, state5_exit_exec, "Reserve")
			{
			}


		/** state (Reserve) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (Release) enter executives **/
		FSM_STATE_ENTER_FORCED (6, state6_enter_exec, "Release")
			{
			}


		/** state (Release) exit executives **/
		FSM_STATE_EXIT_FORCED (6, state6_exit_exec, "Release")
			{
			}


		/** state (Release) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (EndSim) enter executives **/
		FSM_STATE_ENTER_UNFORCED (7, state7_enter_exec, "EndSim")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (15,HostRsvpDeamon)


		/** state (EndSim) exit executives **/
		FSM_STATE_EXIT_UNFORCED (7, state7_exit_exec, "EndSim")
			{
			}


		/** state (EndSim) transition processing **/
		FSM_TRANSIT_MISSING ("EndSim")
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,HostRsvpDeamon)
	}




void
HostRsvpDeamon_svar (prs_ptr,var_name,var_p_ptr)
	HostRsvpDeamon_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (HostRsvpDeamon_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("SessionCallFlag" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_SessionCallFlag);
	if (Vos_String_Equal ("SenderCallFlag" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_SenderCallFlag);
	if (Vos_String_Equal ("ReleaseCallFlag" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ReleaseCallFlag);
	if (Vos_String_Equal ("ReserveCallFlag" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ReserveCallFlag);
	if (Vos_String_Equal ("RecDataPacket" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RecDataPacket);
	if (Vos_String_Equal ("RsvpUpCalls" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RsvpUpCalls);
	if (Vos_String_Equal ("SessionId" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_SessionId);
	if (Vos_String_Equal ("GrpId" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_GrpId);
	if (Vos_String_Equal ("GrpNo" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_GrpNo);
	if (Vos_String_Equal ("LinkNo" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_LinkNo);
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("InStrm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_InStrm);
	if (Vos_String_Equal ("data_length" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_data_length);
	if (Vos_String_Equal ("RecvPkptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RecvPkptr);
	if (Vos_String_Equal ("NoOfLinks" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfLinks);
	if (Vos_String_Equal ("HostIPAddress" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_HostIPAddress);
	if (Vos_String_Equal ("LocalAddress" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_LocalAddress);
	if (Vos_String_Equal ("rsvpSession" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rsvpSession);
	if (Vos_String_Equal ("rsvpSender" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rsvpSender);
	if (Vos_String_Equal ("rsvpReserve" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_rsvpReserve);
	if (Vos_String_Equal ("PSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_PSBList);
	if (Vos_String_Equal ("RSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RSBList);
	if (Vos_String_Equal ("TCSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_TCSBList);

	FOUT;
	}




void
HostRsvpDeamon_diag ()
	{
	Packet* ReceivedPkptr;
	Packet* UpcallPkptr;
	Packet* ApplPkptr;

	FIN (HostRsvpDeamon_diag ())


	FOUT;
	}




void
HostRsvpDeamon_terminate ()
	{
	Packet* ReceivedPkptr;
	Packet* UpcallPkptr;
	Packet* ApplPkptr;

	FIN (HostRsvpDeamon_terminate ())


	FOUT;
	}



Compcode
HostRsvpDeamon_init (pr_state_pptr)
	HostRsvpDeamon_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (HostRsvpDeamon_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (HostRsvpDeamon)",
			sizeof (HostRsvpDeamon_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (HostRsvpDeamon_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



	
	void
	FillCommonHdr( CommonHdr * ch, int msgtype, int ttl )
	{
		ch->Msg_Type = msgtype;
		ch->Send_TTL = ttl;
		ch->Version = RSVP_CURRENT_VER;
	}
	
	void
	FillSession( SessionObj *ses, int sid )
	{
		ses->DestAddress = rsvpSession[sid].DestAddress;
		ses->Protocol_Id = rsvpSession[sid].ProtocolId;
		ses->DestPort    = rsvpSession[sid].DestPort;
	}
	
	int 
	Session( IPAddress da, int pid, int dp, void (*Up)() )
	{
		int i;
		static int firstFlag=0;
	
		if( !firstFlag )
		{
			firstFlag=1;
			memset( rsvpSession, 0, sizeof(RsvpSession)*MAXNOOFSESSIONS );
		}
	
		for(i=1; i<MAXNOOFSESSIONS; ++i)
		{
			if( rsvpSession[i].isfull == 0 )
			{
				rsvpSession[i].isfull = 1;
				rsvpSession[i].DestAddress = da;
				rsvpSession[i].ProtocolId = pid;
				rsvpSession[i].DestPort = dp;
				rsvpSession[i].UpcallProc = Up;
				VL NOV	printf("Session Information is, DestAddress = %d.%d, DestPort = %d\n", 
						rsvpSession[i].DestAddress.NetNum, rsvpSession[i].DestAddress.NodeNum,
						rsvpSession[i].DestPort );
				return(i);
			}
		}
		return(-1);
	}
	
	void
	MakeSessionCall( Packet* ApplPkptr )
	{
	    int tmpSessionId;
	    int pid, dest_port;
	    int dest_net, dest_node;
		IPAddress da;
		Packet* UpcallPkptr;
	 
	    op_pk_nfd_get( ApplPkptr, "DestNet", &dest_net );
	    op_pk_nfd_get( ApplPkptr, "DestNode", &dest_node );
	    op_pk_nfd_get( ApplPkptr, "DestPort", &dest_port );
	 
		da.NetNum = dest_net;
		da.NodeNum = dest_node;
	    tmpSessionId = Session( da, pid, dest_port, NULL );
	
	 
	    UpcallPkptr = op_pk_create_fmt("UPCALL_SESSION");
	    op_pk_nfd_set( UpcallPkptr, "UpcallType", SESSION_MSG );
	    op_pk_nfd_set( UpcallPkptr, "SessionId", tmpSessionId );
	 
	    op_pk_send( UpcallPkptr, APPL_STREAM );
	
		/*  VL	printf("Session Id that was passed to UPcall is %d\n", tmpSessionId);  */
	}
	
	/*
	void 
	Sender( int sid, IPAddress sa, int sp, SenderTemplateObj sto,
			SenderTspecObj sts, int ttl )
	{
		int i;
		PathMessage *pmEle;
		static int firstFlag=0;
	
		if( sid < 1 )
		{
			return;
		}
	
		if( !firstFlag )
		{
			firstFlag=1;
			memset( rsvpSender, 0, sizeof(RsvpSender)*MAXNOOFSESSIONS );
		}
		 VL printf("Sender address %d.%d\n", sa.NetNum, sa.NodeNum); 
		pmEle = (PathMessage *) op_prg_mem_alloc(sizeof(PathMessage));
		rsvpSender[sid].SessionID = sid;
		rsvpSender[sid].SrcAddress = sa;
		rsvpSender[sid].SrcPort = sp;
		rsvpSender[sid].Sender_Template_Object = sto;
		rsvpSender[sid].Sender_Tspec_Object = sts;
		rsvpSender[sid].TTL = ttl;
		FillCommonHdr( &pmEle.common_header, PATH_MESSAGE, rsvpSender[sid].TTL );
		FillSession( &pmEle.Session_Object, sid );
		pmEle.Rsvp_Hop_Object.Next_Hop_Address = sa;
		pmEle.Rsvp_Hop_Object.Logical_Interface_Handle = sa;
		pmEle.Time_Values_Object.Refresh_Period = REFRESH_PERIOD;
		pmEle.Sender_Descriptor.Sender_Template_Object = 
						rsvpSender[sid].Sender_Template_Object;
		pmEle.Sender_Descriptor.Sender_Tspec_Object = 
						rsvpSender[sid].Sender_Tspec_Object;
		VL printf("%d.%d \n", LocalAddress[0].NetNum,LocalAddress[0].NodeNum);
		SendPktToIP( (char *) &pmEle, PATH_MESSAGE ); 
		op_prg_mem_free(pmEle);
	}
	*/
	void
	MakeSenderCall( Packet* ApplPkptr )
	{
		int i, sid, src_net, src_node, src_port, send_ttl;
		PathMessage * PMEle;
		IPAddress sa;
		static int firstFlag=0;
		SenderTemplateObj sto;
		SenderTspecObj sts;
		Packet* ipdgramptr;
		Packet* PathMsgptr;
		int data_len;
	
		PMEle = (PathMessage *) op_prg_mem_alloc( sizeof(PathMessage) );
		memset( PMEle, 0, sizeof(PathMessage) );
		memset( &sto, 0, sizeof(SenderTemplateObj) );
		memset( &sts, 0, sizeof(SenderTspecObj) );
	
		op_pk_nfd_get( ApplPkptr, "SessionId", &sid );
		op_pk_nfd_get( ApplPkptr, "SourceNet", &src_net );
		op_pk_nfd_get( ApplPkptr, "SourceNode", &src_node );
		op_pk_nfd_get( ApplPkptr, "SourcePort", &src_port );
		op_pk_nfd_get( ApplPkptr, "ttl", &send_ttl );
		op_pk_nfd_get( ApplPkptr, "BandWidth", &sts.req_bandwidth );
		op_pk_nfd_get( ApplPkptr, "Delay", &sts.link_delay );
	
		sa = LocalAddress[0];
		sto.SrcAddress = sa;
		sto.SrcPort = src_port;
		/*  VL	printf("Received a Sender call with sid %d\n", sid);  */
		rsvpSender[sid].SessionID = sid;
	    rsvpSender[sid].SrcAddress = sa;
	    rsvpSender[sid].SrcPort = src_port;
	    op_prg_mem_copy( &sto, &(rsvpSender[sid].Sender_Template_Object), 
						 sizeof(SenderTemplateObj) );
	    op_prg_mem_copy( &sts, &(rsvpSender[sid].Sender_Tspec_Object),
						 sizeof(SenderTspecObj) );
	    rsvpSender[sid].TTL = send_ttl;
	
		/** Create Path Message **/
		/* Common Header */
		PMEle->common_header.Version = RSVP_CURRENT_VER;
		PMEle->common_header.Msg_Type = PATH_MESSAGE;
		PMEle->common_header.Send_TTL = send_ttl;
		/* Session Object */
		FillSession( &(PMEle->Session_Object), sid );
		PMEle->Rsvp_Hop_Object.Next_Hop_Address = sa;
		PMEle->Rsvp_Hop_Object.Logical_Interface_Handle = sa;
		PMEle->Time_Values_Object.Refresh_Period = REFRESH_PERIOD;
		/*
		op_prg_mem_copy( &sto, &(PMEle->Sender_Descriptor.Sender_Template_Object),
	                     sizeof(SenderTemplateObj) );
		op_prg_mem_copy( &sts, &(PMEle->Sender_Descriptor.Sender_Tspec_Object),
	                     sizeof(SenderTspecObj) );
		*/
		
		PMEle->Sender_Descriptor.Sender_Template_Object = sto;
		PMEle->Sender_Descriptor.Sender_Tspec_Object = sts;
	
	
		/** Send Path Message **/
		SendPktToIP( (char *) PMEle, 0 , PATH_MESSAGE );
		/*
			PathMsgptr = op_pk_create_fmt("Path_pkfmt");
			op_pk_nfd_set( PathMsgptr, "PathMsg", PMEle, op_prg_mem_alloc, 
						   op_prg_mem_free, sizeof(PathMessage) );
			op_pk_nfd_set( PathMsgptr, "Msg_Type", PATH_MESSAGE );
		 
			data_len = op_pk_total_size_get(PathMsgptr);
			ipdgramptr = op_pk_create_fmt("ip_dgram");
			op_pk_nfd_set( ipdgramptr, "data", PathMsgptr );
			op_pk_nfd_set( ipdgramptr, "prot_id", RSVP_PROTOCOL_ID );
			op_pk_bulk_size_set( ipdgramptr, data_len*8 );
			op_pk_nfd_set( ipdgramptr, "dest_net", PMEle->Session_Object.DestAddress.NetNum );
			op_pk_nfd_set( ipdgramptr, "dest_node", PMEle->Session_Object.DestAddress.NodeNum );
			op_pk_nfd_set( ipdgramptr, "ttl", PMEle->common_header.Send_TTL );
			op_pk_nfd_set( ipdgramptr, "src_net", PMEle->Rsvp_Hop_Object.Next_Hop_Address.NetNum ); 
			op_pk_nfd_set( ipdgramptr, "src_node", PMEle->Rsvp_Hop_Object.Next_Hop_Address.NodeNum );
			op_pk_nfd_set( ipdgramptr, "orig_len", data_len );
			op_pk_nfd_set( ipdgramptr, "frag_len", data_len );
			op_pk_nfd_set( ipdgramptr, "frag", 0 );
			op_pk_send( ipdgramptr, IPSTREAM );
			VL printf("Sent the PATH message from %d.%d to %d.%d\n", 
					PMEle->Rsvp_Hop_Object.Next_Hop_Address.NetNum,
					PMEle->Rsvp_Hop_Object.Next_Hop_Address.NodeNum,
					PMEle->Session_Object.DestAddress.NetNum, 
					PMEle->Session_Object.DestAddress.NodeNum );
		*/
	}
	
	void
	Reserve( int sid, FlowDescriptorList * fdl )
	{
		int i;
		static int firstFlag=0;
		ResvMessage rmEle;
	
		if( sid < 1 )
		{
			return;
		}
	
		if( !firstFlag )
		{
			firstFlag=1;
			memset( rsvpReserve, 0, sizeof(RsvpReserve)*MAXNOOFSESSIONS );
		}
		for( i=0; i<InLIST_SIZE; ++i )
		{
			rmEle.flow_descriptor_list[i] = fdl[i];
		}
		FillCommonHdr( &rmEle.common_header, RESV_MESSAGE, 0 );
		FillSession( &rmEle.Session_Object, sid );
		/* OPNET CALL */
		/* Send the resv message */
	}
	
	void
	MakeReserveCall( Packet* ApplPkptr)
	{
		int sid;
		static int firstFlag=0;
		ResvMessage * rmEle;
	
		FlowSpecObj fs;
		FilterSpecObj fls;
	
	    if( !firstFlag )
	    {
	        firstFlag=1;
	        memset( rsvpReserve, 0, sizeof(RsvpReserve)*MAXNOOFSESSIONS );
		}
	
		rmEle = (ResvMessage *)calloc( 1,sizeof(ResvMessage) );
		op_pk_nfd_get( ApplPkptr, "SessionId", &sid );
		if( sid < 1 )
	    {
	        return;
	    }
	 
		op_pk_nfd_get( ApplPkptr, "SenderNet", &fls.SrcAddress.NetNum );
		op_pk_nfd_get( ApplPkptr, "SenderNode", &fls.SrcAddress.NodeNum );
		op_pk_nfd_get( ApplPkptr, "SenderPort", &fls.SrcPort );
		op_pk_nfd_get( ApplPkptr, "BandWidth", &fs.req_bandwidth );
		op_pk_nfd_get( ApplPkptr, "Delay", &fs.link_delay );
	
		rmEle->flow_descriptor_list[0].Flowspec_Object = fs;
		rmEle->flow_descriptor_list[0].Filter_Spec_Object = fls;
		
		/*  VL printf("Sending the Reserve Message \n");  */
		FillCommonHdr( &rmEle->common_header, RESV_MESSAGE, DEFAULT_RESERVE_TTL );
	    FillSession( &rmEle->Session_Object, sid );
		rmEle->Rsvp_Hop_Object.Logical_Interface_Handle.NetNum=LocalAddress[0].NetNum;
		rmEle->Rsvp_Hop_Object.Logical_Interface_Handle.NodeNum=5;
	
		SendResvPktToIP( (char *) rmEle, rmEle->Rsvp_Hop_Object.Logical_Interface_Handle );
	}
	 
	void
	Release( int sid )
	{
		if( sid < 1 )
			return;
		if( rsvpSender[sid].SessionID == sid )
		{
			/* Path Tear */
		}
		else if( rsvpReserve[sid].SessionID == sid )
		{
			/* Resv Tear */
		}
		else
		{
			return;
		}
	}
	
	void
	HostInitialize()
	{
		char mess[100];
		int i;
		Objid sub_netid;
		/* Initialize LocalAddress */
		my_id = op_id_self();
		ListInit(&PSBList);
		ListInit(&RSBList);
		ListInit(&TCSBList);
	
	
		ip_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE , 0 );
	
	    for( i=0; i<1; ++i )
	    {
	        sprintf( mess,"net_num%1d",i );
	        op_ima_obj_attr_get( ip_objid, mess, &LocalAddress[i].NetNum );
	        sprintf( mess,"node_num%1d",i );
	        op_ima_obj_attr_get( ip_objid, mess, &LocalAddress[i].NodeNum );
			if(OFF)	VL	printf("Host LocalAddress %d is %d.%d\n", i, 
						LocalAddress[i].NetNum, LocalAddress[i].NodeNum );
	    }
		/*  InitVariables( LocalAddress, PSBList, RSBList, TCSBList );  */
	}
