/* Process model C form file: RsvpOnHosts.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "RsvpOnHosts.pr.h"
FSM_EXT_DECS



/* Header block */
#include<stdio.h>
#define ARRIVAL op_intrpt_type()==OPC_INTRPT_STRM

#define Path_Msg_Refresh_timeout 4000

#define InLIST_SIZE     		10
#define TCP_PROTOCOL_ID			12
#define UDP_PROTOCOL_ID			20
#define RSVP_PROTOCOL_ID		46
#define REFRESH_PERIOD     		300

#define RSVP_CURRENT_VER		0x01

#define PATH_MESSAGE			0x01
#define RESV_MESSAGE			0x02
#define PATH_ERROR				0x03
#define RESV_ERROR				0x04
#define PATH_TEAR				0x05
#define RESV_TEAR				0x06
#define RESV_CONF				0x07

#define OFF		0 
#define ON		1 
#define UNDEFINED		-1
#define MCAST_ADDRESS 	500
#define REFRESH_PERIOD	300


/*
** IP Address structure
*/
typedef struct {
	int NetNum;
	int NodeNum;
} IPAddress;

/* 
** Session Information
*/
typedef struct {
	IPAddress DestAddress;
	int ProtocolId;
	int DestPort;
} SessInfo;

/*
** Common Header
*/
typedef struct {
	int Version;
	int Msg_Type;
	int Send_TTL;
	double RSVP_Checksum;
	double RSVP_Length;
} CommonHdr;

/*
** Session object 
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress DestAddress;
	int Protocol_Id;
	int Flags;
	int DestPort;
} SessionObj;

/*
** RSVP Hop Object
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress Next_Hop_Address;
	IPAddress Logical_Interface_Handle;
} RsvpHopObj;

/*
** Time values object
*/
typedef struct {
	int Class_Num; 
	double Length;
	double Refresh_Period;
} TimeValObj;

/*
** Sender template
*/
typedef struct {
	int Class_Num; 
	double Length;
	IPAddress SrcAddress;
	int SrcPort;
} SenderTemplateObj;

/*
** Sender T-Spec object
*/
typedef struct {
	int Class_Num; 
	double Length;
	double req_bandwidth;
	double link_delay;
} SenderTspecObj;

/*
** FlowSpec Object
*/
typedef SenderTspecObj FlowSpecObj;

/*
** Filter Spec
*/
typedef SenderTemplateObj FilterSpecObj;

/*
** Style 
*/
typedef struct {
	int Class_Num;
	double Length;
	int Resv_Style;
} StyleObj;

/*
** Scope object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress SrcAddress;
} ScopeObj;

/*
** Reservation Confirm object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress Receiver_Address;
} ResvConfirmObj;

/*
** Error Spec object
*/
typedef struct {
	int Class_Num;
	double Length;
	IPAddress Error_Node_Address;
	int Flags;
	int Error_code;
} ErrorSpecObj;



/* 
** Sender Descriptor structure 
*/
typedef struct {

	/* SENDER_TEMPLATE structure */
  SenderTemplateObj Sender_Template_Object; 

	/* SENDER_TSPEC structure */
  SenderTspecObj Sender_Tspec_Object;     

}SenderDescriptor;

/* 
** Flow descriptor list structure 
*/
typedef struct {

	/* FLOWSPEC structure */
  FlowSpecObj Flowspec_Object;         

  /* this was added when it was known that FF style needs to be considered */
	/* FILTER_SPEC structure */
  FilterSpecObj Filter_Spec_Object;      

}FlowDescriptorList;

/* 
** PATH message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header;

	/* SESSION object structure */
  SessionObj Session_Object;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object;

	/* TIME_VALUES Object structure */
  TimeValObj Time_Values_Object;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor;

}PathMessage;

/* 
** RESV message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* TIME_VALUES Object structure */
  TimeValObj Time_Values_Object       ;

	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list[InLIST_SIZE];

}ResvMessage;

/* 
** PathTear message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor        ;

}PathTearMessage;

/* 
** ResvTear message structure 
*/
typedef struct {
	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* RSVP_HOP Object structure */
  RsvpHopObj Rsvp_Hop_Object          ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list  ;

}ResvTearMessage;

/* 
** PathErr message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;

	/* Sender Descriptor structure */
  SenderDescriptor Sender_Descriptor        ;

}PathErrMessage;

/* 
** ResvErr message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;

	/* SCOPE Object structure */
  ScopeObj Scope_Object;

	/* STYLE Object structure */
  StyleObj Style_Object;

	/* Flow descriptor list structure */
  /*  ErrorFlowDescriptor error_flow_descriptor;  */

} ResvErrMessage;

/* 
** ResvConf message structure 
*/
typedef struct {

	/* Common Header structure */
  CommonHdr common_header            ;

	/* SESSION object structure */
  SessionObj Session_Object           ;

	/* ERROR_SPEC Object structure */
  ErrorSpecObj Error_Spec_Object        ;


	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Flow descriptor list structure */
  FlowDescriptorList flow_descriptor_list  ;

}ResvConfMessage;

/* 
** Path State Block (PSB) structure 
*/
typedef struct {

	/* SESSION object structure */
  SessionObj Session_Object;          

	/* SENDER_TEMPLATE structure */
  SenderTemplateObj Sender_Template_Object;  

	/* SENDER_TSPEC structure */
  SenderTspecObj Sender_Tspec_Object;     

	/* Previous Hop IP address from PHOP object */
  IPAddress Previous_Hop_Address;       

	/* Logical Interface Handle from PHOP object*/
  IPAddress LIH;                     

	/* Remaining IP TTL */
  int rem_IP_TTL;                 

	/* To determine EDGE of the RSVP network */
  int E_Police_Flag;              

	/* To determine if the App. is local or not */
  int Local_Only_Flag;            

	/* List of outgoing Interfaces for this 
	   (sender, destination) */
  IPAddress OutInterface_List[InLIST_SIZE];        

	/* Expected incoming interface */
  IPAddress IncInterface                ;

	/* Blockade FLOWSPEC */
  FlowSpecObj Qb;         

}PathStateBlock;

/* 
** Reservation State Block (RSB) structure 
*/
typedef struct {

	/* SESSION object structure */
  SessionObj Session_Object;          

	/* Next Hop IP address from PHOP object */
  IPAddress Next_Hop_Address;           

	/* Outgoing Interface on which reservation is 
	   to be made or has been made*/
  IPAddress OI;                      

	/* Newly added when it was known that FF style need to be considered */
	/* This would be a single FILTER_SPEC for FF style */
  FilterSpecObj Filter_Spec_Object;  

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* FLOWSPEC structure */
  FlowSpecObj Flowspec_Object;         

	/* SCOPE Object structure */
  ScopeObj Scope_Object             ;

	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object      ;

}ResvStateBlock;

/* 
** Traffic Control State Block 
*/
typedef struct {

	/* SESSION object structure */
  SessionObj Session_Object;          

	/* Outgoing Interface on which reservation is 
	   to be made or has been made*/
  IPAddress OI;                      

	/* Newly added when it was known that FF style need to be considered */
  FilterSpecObj Filter_Spec_Object[InLIST_SIZE];  

	/* STYLE Object structure */
  StyleObj Style_Object             ;

	/* Effective flowspec (LUB of FLOWSPEC's from 
       matching RSB's) structure */
  FlowSpecObj TC_Flowspec;        

	/* Updated object structure to be forwarded after merging */
  FlowSpecObj Fwd_Flowspec;       

	/* Effective sender Tspec structure */
  SenderTspecObj TC_Tspec;        

	/* Entry Police flags */
  int E_Police_Flag;              

	/* Merge Police flags: Need not use this one*/
  int M_Police_Flag;              

	/* Branch Police flags */
  int B_Police_Flag;              

	/* Reservation Handle */
  int Rhandle;                    

	/* Filter handle list */
  int * F_Handle_list;            

	/* RESV_CONFIRM Object structure */
  ResvConfirmObj Resv_Confirm_Object;

}TrafficControlStateBlock;


#define PrintErr(x)   printf("%s : Filename %s : Line NO %d:\n", \
								x, __FILE__, __LINE__);

struct GenList {
	struct GenList * next;
	struct GenList * curr;
 	char *ele;
};

typedef struct GenList GenList;

/*
** Function Prototypes
*/
int CmpSession(SessionObj * so1, SessionObj *so2);
int CmpSenderTspec(SenderTspecObj * sts1, SenderTspecObj *sts2);
int CmpSenderTemplate(SenderTemplateObj * st1, SenderTemplateObj *st2);
int CmpFlowSpec(FlowSpecObj * sts1, FlowSpecObj *sts2);
int CmpFilterSpec(FilterSpecObj * fs1, FilterSpecObj *fs2);
int CmpScope(ScopeObj * sc1, ScopeObj *sc2);
int CmpAddress(IPAddress * ipa1, IPAddress * ipa2);
int UndefAddress( IPAddress * ipa );
int IsMcast( IPAddress * ipa );
int CmpAddressL(IPAddress * ipa1, IPAddress * ipa2);
int CmpLocalAddress(IPAddress * ipa);
void MakeAddress( IPAddress * ipa, int netno, int nodeno);
void CpyAddress(IPAddress * ipa1, IPAddress * ipa2);
void RouterInitialize();
void InitAddressL( IPAddress * ipa );
void CpyAddressL(IPAddress * ipa1, IPAddress * ipa2);
void PathMsgPro(PathMessage *pmess, IPAddress InIf);
void ResvMsgPro(ResvMessage *rmess);
void UpdateTrafficControl(ResvStateBlock *activeRSB);
void ResvRefresh( ResvStateBlock *rsbEle );
void PathTearPro(PathTearMessage *pmess);
void DeleteListEle( GenList *lst, char *Ele, int size );
void ListInit(GenList* lst);
void * AddListEle( GenList *lst, char *Ele, int size );
char * GetNextListEle( GenList *lst, int flag );
int PathRefresh( PathStateBlock *psbEle , IPAddress OI);
int IsListEmpty(GenList* lst);



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_id;
	Objid                   		sv_ip_objid;
	int                     		sv_NoOfLinks;
	int                     		sv_in_strm;
	int                     		sv_ip_strm;
	struct GenList          		sv_PSBList;
	struct GenList          		sv_RSBList;
	struct GenList          		sv_TCSBList;
	IPAddress               		sv_LocalAddress[InLIST_SIZE];
	int                     		sv_obt_pk_srcnet;
	int                     		sv_obt_pk_srcnode;
	int                     		sv_send_pathmsg;
	int                     		sv_send_resvmsg;
	int                     		sv_send_pathtearmsg;
	int                     		sv_send_resvtearmsg;
	int                     		sv_send_resvconf;
	int                     		sv_protocolid;
	int                     		sv_Indep_Rou;
	} RsvpOnHosts_state;

#define pr_state_ptr            		((RsvpOnHosts_state*) SimI_Mod_State_Ptr)
#define my_id                   		pr_state_ptr->sv_my_id
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define NoOfLinks               		pr_state_ptr->sv_NoOfLinks
#define in_strm                 		pr_state_ptr->sv_in_strm
#define ip_strm                 		pr_state_ptr->sv_ip_strm
#define PSBList                 		pr_state_ptr->sv_PSBList
#define RSBList                 		pr_state_ptr->sv_RSBList
#define TCSBList                		pr_state_ptr->sv_TCSBList
#define LocalAddress            		pr_state_ptr->sv_LocalAddress
#define obt_pk_srcnet           		pr_state_ptr->sv_obt_pk_srcnet
#define obt_pk_srcnode          		pr_state_ptr->sv_obt_pk_srcnode
#define send_pathmsg            		pr_state_ptr->sv_send_pathmsg
#define send_resvmsg            		pr_state_ptr->sv_send_resvmsg
#define send_pathtearmsg        		pr_state_ptr->sv_send_pathtearmsg
#define send_resvtearmsg        		pr_state_ptr->sv_send_resvtearmsg
#define send_resvconf           		pr_state_ptr->sv_send_resvconf
#define protocolid              		pr_state_ptr->sv_protocolid
#define Indep_Rou               		pr_state_ptr->sv_Indep_Rou




/* Process model interrupt handling procedure */



void
RsvpOnHosts ()
	{


	FSM_ENTER (RsvpOnHosts)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init")
			{
			RouterInitialize();
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init")
			{
			}


		/** state (init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,RsvpOnHosts)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle")
			{
			}


		/** state (idle) transition processing **/
		FSM_TRANSIT_ONLY ((ARRIVAL), 2, state2_enter_exec, ;, "idle")
		/*---------------------------------------------------------*/



		/** state (arr) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "arr")
			{
			{
			Packet *ReceivedPkptr;
			Packet *ReceivedRsvpPkptr;
			int ReceivedMsgType; 
			
			in_strm = op_intrpt_strm();
			 
			ReceivedPkptr = op_pk_get( in_strm );
			 
			op_pk_nfd_get( ReceivedPkptr, "strm", &ip_strm );
			op_pk_nfd_get( ReceivedPkptr, "prot_id", &protocolid );
			op_pk_nfd_get( ReceivedPkptr, "src_net", &obt_pk_srcnet );
			op_pk_nfd_get( ReceivedPkptr, "src_node", &obt_pk_srcnode );
			 
			printf( "A Packet is obtained from the router %d.%d through the \
			        stream %d with the protocol Id of %d \n", obt_pk_srcnet,
			        obt_pk_srcnode, ip_strm, protocolid );
			 
			switch(protocolid) 
			{
					case 46  :  
				op_pk_nfd_get(ReceivedPkptr, "data", &ReceivedRsvpPkptr);
				op_pk_nfd_get(ReceivedRsvpPkptr, "Msg_Type", &ReceivedMsgType);
				switch( ReceivedMsgType )
				{
					case PATH_MESSAGE:
						PathMsgPro( (PathMessage *)ReceivedRsvpPkptr, LocalAddress[in_strm] );
					break;
					case RESV_MESSAGE:
						ResvMsgPro( (ResvMessage *)ReceivedRsvpPkptr );
					break;
					case PATH_ERROR:
					break;
					case RESV_ERROR:
					break;
					case PATH_TEAR:
						PathTearPro( (PathTearMessage *)ReceivedRsvpPkptr );
					break;
					case RESV_TEAR:
					break;
					case RESV_CONF:
					break;
				}
				break;
			}
			 
			op_pk_destroy( ReceivedPkptr );
			 
			}
			}


		/** state (arr) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "arr")
			{
			}


		/** state (arr) transition processing **/
		FSM_INIT_COND (send_pathmsg)
		FSM_TEST_COND (send_resvmsg)
		FSM_TEST_COND (send_pathtearmsg)
		FSM_TEST_COND (send_resvtearmsg)
		FSM_TEST_COND (send_resvconf)
		FSM_DFLT_COND
		FSM_TEST_LOGIC ("arr")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 3, state3_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 4, state4_enter_exec, ;)
			FSM_CASE_TRANSIT (2, 5, state5_enter_exec, ;)
			FSM_CASE_TRANSIT (3, 6, state6_enter_exec, ;)
			FSM_CASE_TRANSIT (4, 7, state7_enter_exec, ;)
			FSM_CASE_TRANSIT (5, 1, state1_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (pathmsg) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "pathmsg")
			{
			}


		/** state (pathmsg) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "pathmsg")
			{
			}


		/** state (pathmsg) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (resvmsg) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "resvmsg")
			{
			}


		/** state (resvmsg) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "resvmsg")
			{
			}


		/** state (resvmsg) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (ptearmsg) enter executives **/
		FSM_STATE_ENTER_FORCED (5, state5_enter_exec, "ptearmsg")
			{
			}


		/** state (ptearmsg) exit executives **/
		FSM_STATE_EXIT_FORCED (5, state5_exit_exec, "ptearmsg")
			{
			}


		/** state (ptearmsg) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (rtearmsg) enter executives **/
		FSM_STATE_ENTER_FORCED (6, state6_enter_exec, "rtearmsg")
			{
			}


		/** state (rtearmsg) exit executives **/
		FSM_STATE_EXIT_FORCED (6, state6_exit_exec, "rtearmsg")
			{
			}


		/** state (rtearmsg) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (rconfmsg) enter executives **/
		FSM_STATE_ENTER_FORCED (7, state7_enter_exec, "rconfmsg")
			{
			}


		/** state (rconfmsg) exit executives **/
		FSM_STATE_EXIT_FORCED (7, state7_exit_exec, "rconfmsg")
			{
			}


		/** state (rconfmsg) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,RsvpOnHosts)
	}




void
RsvpOnHosts_svar (prs_ptr,var_name,var_p_ptr)
	RsvpOnHosts_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (RsvpOnHosts_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("NoOfLinks" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfLinks);
	if (Vos_String_Equal ("in_strm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_in_strm);
	if (Vos_String_Equal ("ip_strm" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_strm);
	if (Vos_String_Equal ("PSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_PSBList);
	if (Vos_String_Equal ("RSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RSBList);
	if (Vos_String_Equal ("TCSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_TCSBList);
	if (Vos_String_Equal ("LocalAddress" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_LocalAddress);
	if (Vos_String_Equal ("obt_pk_srcnet" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_obt_pk_srcnet);
	if (Vos_String_Equal ("obt_pk_srcnode" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_obt_pk_srcnode);
	if (Vos_String_Equal ("send_pathmsg" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_pathmsg);
	if (Vos_String_Equal ("send_resvmsg" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_resvmsg);
	if (Vos_String_Equal ("send_pathtearmsg" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_pathtearmsg);
	if (Vos_String_Equal ("send_resvtearmsg" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_resvtearmsg);
	if (Vos_String_Equal ("send_resvconf" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_resvconf);
	if (Vos_String_Equal ("protocolid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_protocolid);
	if (Vos_String_Equal ("Indep_Rou" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_Indep_Rou);

	FOUT;
	}




void
RsvpOnHosts_diag ()
	{

	FIN (RsvpOnHosts_diag ())


	FOUT;
	}




void
RsvpOnHosts_terminate ()
	{

	FIN (RsvpOnHosts_terminate ())


	FOUT;
	}



Compcode
RsvpOnHosts_init (pr_state_pptr)
	RsvpOnHosts_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (RsvpOnHosts_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (RsvpOnHosts)",
			sizeof (RsvpOnHosts_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (RsvpOnHosts_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



	int 
	CmpSession(SessionObj * so1, SessionObj *so2)
	{
		if( so1->DestAddress.NetNum == so2->DestAddress.NetNum &&
		    so1->DestAddress.NodeNum == so2->DestAddress.NodeNum &&
			so1->Protocol_Id == so2->Protocol_Id &&
			so1->DestPort == so2->DestPort )
		return 1;
		return 0;
	}
	
	int 
	CmpSenderTspec(SenderTspecObj * sts1, SenderTspecObj *sts2)
	{
		if( sts1->req_bandwidth == sts2->req_bandwidth &&
			sts1->link_delay == sts2->link_delay )
		return 1;
		return 0;
	}
	
	int 
	CmpSenderTemplate(SenderTemplateObj * st1, SenderTemplateObj *st2)
	{
		if( st1->SrcAddress.NetNum == st2->SrcAddress.NetNum &&
		    st1->SrcAddress.NodeNum == st2->SrcAddress.NodeNum &&
			st1->SrcPort == st2->SrcPort )
		return 1;
		return 0;
	}
	
	int 
	CmpFlowSpec(FlowSpecObj * sts1, FlowSpecObj *sts2)
	{
		if( sts1->req_bandwidth == sts2->req_bandwidth &&
			sts1->link_delay == sts2->link_delay )
		return 1;
		return 0;
	}
	
	int 
	CmpFilterSpec(FilterSpecObj * fs1, FilterSpecObj *fs2)
	{
		if( fs1->SrcAddress.NetNum == fs2->SrcAddress.NetNum &&
		    fs1->SrcAddress.NodeNum == fs2->SrcAddress.NodeNum &&
			fs1->SrcPort == fs2->SrcPort )
		return 1;
		return 0;
	}
	
	int 
	CmpScope(ScopeObj * sc1, ScopeObj *sc2)
	{
		if( sc1->SrcAddress.NetNum == sc2->SrcAddress.NetNum &&
		    sc1->SrcAddress.NodeNum == sc2->SrcAddress.NodeNum )
		return 1;
		return 0;
	}
	
	int 
	CmpAddress(IPAddress * ipa1, IPAddress * ipa2)
	{
		if( ipa1->NetNum == ipa2->NetNum &&
		    ipa1->NodeNum == ipa2->NodeNum )
		return 1;
		return 0;
			
	}
	
	void
	MakeAddress( IPAddress * ipa, int netno, int nodeno)
	{
		ipa->NetNum = netno;
		ipa->NodeNum = nodeno;
	}
	
	void
	CpyAddress(IPAddress * ipa1, IPAddress * ipa2)
	{
		ipa1->NetNum = ipa2->NetNum;
		ipa1->NodeNum = ipa2->NodeNum;
	}
	
	void
	RouterInitialize()
	{
		char mess[100];
		int i;
		Objid sub_netid;
		ListInit(&PSBList);
		ListInit(&RSBList);
		ListInit(&TCSBList);
		/* Initialize LocalAddress */
		my_id = op_id_self();
	
		ip_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 0 );
	
	
		op_ima_obj_attr_get( ip_objid, "Indep_Rou", &Indep_Rou );
	
		if( !Indep_Rou )
			sub_netid=op_topo_parent( op_topo_parent(my_id));
		else
			sub_netid=op_topo_parent( my_id );
	
		NoOfLinks = op_topo_assoc_count( sub_netid,OPC_TOPO_ASSOC_IN,OPC_OBJTYPE_LKSIMP );
	
		for( i=0; i<NoOfLinks; ++i )
		{
			sprintf( mess,"net_num%1d",i );
			op_ima_obj_attr_get( ip_objid, mess, &LocalAddress[i].NetNum );
			sprintf( mess,"node_num%1d",i );
			op_ima_obj_attr_get( ip_objid, mess, &LocalAddress[i].NodeNum );
		}
	}
	
	int
	UndefAddress( IPAddress * ipa )
	{
		if(ipa->NetNum == 0 || ipa->NetNum == UNDEFINED)
			return 1;
		if(ipa->NodeNum == UNDEFINED)
			return 1;
		return 0;
	}
	
	int
	IsMcast( IPAddress * ipa )
	{
		if( ipa->NetNum == MCAST_ADDRESS )
			return 1;
		return 0;
	}
	
	void
	InitAddressL( IPAddress * ipa )
	{
		int i;
		for( i=0; i<InLIST_SIZE; ++i )
		{
			ipa->NetNum = 0;
			ipa->NodeNum = 0;
		}
	}
	
	void 
	CpyAddressL(IPAddress * ipa1, IPAddress * ipa2)
	{
		int i;
		for( i=0; i<InLIST_SIZE; ++i )
		{
			ipa1[i].NetNum = ipa2[i].NetNum;
			ipa1[i].NodeNum = ipa2[i].NodeNum;
		}
	}
	
	int 
	CmpAddressL(IPAddress * ipa1, IPAddress * ipa2)
	{
		int i;
		for( i=0; i<InLIST_SIZE; ++i )
		{
			if( ipa1[i].NetNum != ipa2[i].NetNum &&
				ipa1[i].NodeNum != ipa2[i].NodeNum )
				return 0;
		}
		return 1;
	}
	
	int 
	CmpLocalAddress(IPAddress * ipa)
	{
		int i;
		for( i=0; i<InLIST_SIZE; ++i )
		{
			if( LocalAddress[i].NetNum != ipa->NetNum &&
				LocalAddress[i].NodeNum != ipa->NodeNum )
				return 0;
		}
		return 1;
	}
	
	void
	PathMsgPro(PathMessage *pmess, IPAddress InIf)
	{
		PathStateBlock * fPSB = NULL;
		PathStateBlock * cPSB = NULL;
	
		PathStateBlock * psbEle = NULL;
		ResvStateBlock * rsbEle = NULL;
		ResvStateBlock * activeRSB = NULL;
		int Path_Refresh_Needed_Flag = OFF;
		int Resv_Refresh_Needed_Flag = OFF;
		int foundFlag = OFF;
		int routeChangeNoti;
		int LocalInFlag = OFF;
		int i;
		IPAddress Rt_InIf, Rt_OutL[InLIST_SIZE];
	
		foundFlag = OFF;
		psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 1 );
		while( psbEle )
		{
			if( CmpAddress( &psbEle->Session_Object.DestAddress, 
							&pmess->Session_Object.DestAddress) &&
			    psbEle->Session_Object.Protocol_Id == 
				pmess->Session_Object.Protocol_Id  &&
			    psbEle->Session_Object.DestPort != 
				pmess->Session_Object.DestPort && (
			    psbEle->Session_Object.DestPort == 0 ||
				pmess->Session_Object.DestPort == 0 ))
			{
				/**
				*** Send Path Error Message : Conflicting Destination Port
				**/
				return;
			}
			if( CmpAddress( &psbEle->Sender_Template_Object.SrcAddress,
					&pmess->Sender_Descriptor.Sender_Template_Object.SrcAddress) &&
				psbEle->Sender_Template_Object.SrcPort !=
				pmess->Sender_Descriptor.Sender_Template_Object.SrcPort && (
				psbEle->Sender_Template_Object.SrcPort == 0 ||
				pmess->Sender_Descriptor.Sender_Template_Object.SrcPort==0 ))
			{
				/**
				*** Send Path Error Message : Ambiguous Path
				**/
				return;
			}
			if( CmpSession( &psbEle->Session_Object, &pmess->Session_Object) &&
				CmpSenderTemplate( &psbEle->Sender_Template_Object, 
					&pmess->Sender_Descriptor.Sender_Template_Object) &&
				psbEle->Local_Only_Flag == OFF )
			{
				fPSB = psbEle;
			}
			else
			{
				fPSB = NULL;
			}
			
			/**
			*** Search for the PSB whose session and sender template
			*** matches the corresponding objects in the path message.
			*** and whose incomming interface matches the IncInterface.
			**/
			if( CmpSession( &psbEle->Session_Object, &pmess->Session_Object) &&
				CmpSenderTemplate( &psbEle->Sender_Template_Object, 
					&pmess->Sender_Descriptor.Sender_Template_Object) &&
				CmpAddress( &psbEle->IncInterface, &InIf) )
			{
			/**
			*** If matching PSB is found
			**/
				if( !CmpAddress( &psbEle->Previous_Hop_Address,
					 		&pmess->Rsvp_Hop_Object.Next_Hop_Address) ||
					 !CmpAddress( &psbEle->LIH,
					 		&pmess->Rsvp_Hop_Object.Logical_Interface_Handle) ||
					 !CmpSenderTspec( &psbEle->Sender_Tspec_Object,
						&pmess->Sender_Descriptor.Sender_Tspec_Object) )
				{
					CpyAddress( &psbEle->Previous_Hop_Address,
						 	&pmess->Rsvp_Hop_Object.Next_Hop_Address );
					CpyAddress( &psbEle->LIH,
						 	&pmess->Rsvp_Hop_Object.Logical_Interface_Handle);
					 memcpy( &psbEle->Sender_Tspec_Object,
						&pmess->Sender_Descriptor.Sender_Tspec_Object,
						sizeof(SenderTspecObj) );
					Path_Refresh_Needed_Flag = ON;
				}
				if( !CmpAddress( &psbEle->Previous_Hop_Address,
					 		&pmess->Rsvp_Hop_Object.Next_Hop_Address) ||
					 !CmpAddress( &psbEle->LIH,
					 		&pmess->Rsvp_Hop_Object.Logical_Interface_Handle) )
				{
					Resv_Refresh_Needed_Flag = ON;
				}               
				cPSB = psbEle;
				foundFlag = ON;
				break;
			}               
			psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 0 );
		}
	
		if( foundFlag == OFF )
		{
		/**
		*** If no matching PSB is found
		**/
	
			psbEle = (PathStateBlock *) calloc( 1, sizeof(PathStateBlock) );
			memcpy( &psbEle->Session_Object, 
					&pmess->Session_Object, 
					sizeof(SessionObj));
			memcpy( &psbEle->Sender_Template_Object, 
					&pmess->Sender_Descriptor.Sender_Template_Object,
					sizeof(SenderTemplateObj) );
			memcpy( &psbEle->Sender_Tspec_Object, 
					&pmess->Sender_Descriptor.Sender_Tspec_Object,
					sizeof(SenderTspecObj) );
			CpyAddress( &psbEle->Previous_Hop_Address,
					&pmess->Rsvp_Hop_Object.Next_Hop_Address );
			CpyAddress( &psbEle->LIH, 
					&pmess->Rsvp_Hop_Object.Logical_Interface_Handle );
			if( CmpLocalAddress( 
				&pmess->Sender_Descriptor.Sender_Template_Object.SrcAddress ) )
			{
				CpyAddress( &psbEle->OutInterface_List[0],
					&pmess->Sender_Descriptor.Sender_Template_Object.SrcAddress);
				psbEle->IncInterface.NetNum = UNDEFINED;
				psbEle->IncInterface.NodeNum = UNDEFINED;
				LocalInFlag = ON;
			}
			else
			{
				CpyAddress( &psbEle->IncInterface, &InIf );
				psbEle->Local_Only_Flag = ON;
			}
			AddListEle( &PSBList, (char*)psbEle, sizeof(PathStateBlock) );
			Path_Refresh_Needed_Flag = ON;
			cPSB = psbEle;
		}
	
		if( cPSB == NULL )
		{
			/**
			*** Nothing in cPSB so return
			**/
			return;
		}
	
		/**
		*** Start or Restart the Cleanup timer
		**/
	
		/**
		*** Update cPSB
		**/
		cPSB->rem_IP_TTL = pmess->common_header.Send_TTL;
	
		/**
		*** If the PSB is new or there is no route change notification
		*** but not if the cPSB is from the local API
		**/
	
		/**
		*** routeChangeNoti = 
		**/
		if( (foundFlag == OFF ) && LocalInFlag == OFF )
		{
			if( IsMcast( &cPSB->Session_Object.DestAddress ) )
			{
				Mcast_Route_Query( 
					cPSB->Sender_Template_Object.SrcAddress,
					cPSB->Session_Object.DestAddress, Rt_OutL );
				if( fPSB != cPSB )
				{
					if( fPSB != NULL )
					{
						CpyAddressL( cPSB->OutInterface_List, 
									 fPSB->OutInterface_List );
						InitAddressL(fPSB->OutInterface_List);
						fPSB->Local_Only_Flag = ON;
					}
					cPSB->Local_Only_Flag = OFF;
					fPSB = cPSB;
				}
			}
			if( CmpAddressL( Rt_OutL, fPSB->OutInterface_List ))
			{
				CpyAddressL( fPSB->OutInterface_List, Rt_OutL );
				/**
				*** Execute PATH LOCAL REPAIR
				**/
			}
		}
		if( Path_Refresh_Needed_Flag == OFF )
		{
			return;
		}
		else
		{
			IPAddress tOutList[InLIST_SIZE];
			int olFlag = OFF;
			InitAddressL( tOutList );
			/*  if( localOnlyFlag == OFF )  */
			{
				/**
				*** Make an path event upcall for each local application
				*** for this session
				**/
			}
			/*
			** if outInterfacelist is not empty
			*/
			if( !CmpAddressL( fPSB->OutInterface_List, tOutList )) 
			{
				/**
				*** Execute PATH REFRESH SEQUENCE
				**/
				for(i=0; i<InLIST_SIZE; ++i)  {
					if( !UndefAddress( &fPSB->OutInterface_List[i] ) )
					{
						PathRefresh( fPSB , fPSB->OutInterface_List[i] );
					}
				}
			}
			rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 1 );
			olFlag = OFF;
			while( rsbEle )
			{
				if( CmpFilterSpec( &rsbEle->Filter_Spec_Object,
						(FilterSpecObj *)&fPSB->Sender_Template_Object ) )
				{
					for(i=0; i<InLIST_SIZE; ++i)  {
						if( CmpAddress( &rsbEle->OI, &fPSB->OutInterface_List[i]) )
						{
							activeRSB = rsbEle;
							olFlag = ON;
							break;
						}
					}
				}
				rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 0 );
			}
			if( olFlag == OFF )
			{
				return;
			}
			/**
			*** Execute RESV REFRESH SEQUENCE for the PHOP in the fPSB
			**/
			/**
			*** Execute the UPDATE TRAFFIC CONTROL 
			**/
			return;
		}
	}
	
	int
	PathRefresh( PathStateBlock *psbEle , IPAddress OI)
	{
		PathMessage pm;
		if( psbEle->rem_IP_TTL == 1 )
		{
			/* Life ended */
			return;
		}
		/* TO BE DONE */
		/* E_Police Flag */
		psbEle->rem_IP_TTL--;
		pm.common_header.Send_TTL = psbEle->rem_IP_TTL;
		pm.Time_Values_Object.Refresh_Period = REFRESH_PERIOD;
		memcpy( &pm.Sender_Descriptor.Sender_Template_Object,
				&psbEle->Sender_Template_Object, sizeof(SenderTemplateObj) );
		memcpy( &pm.Sender_Descriptor.Sender_Tspec_Object,
				&psbEle->Sender_Tspec_Object, sizeof(SenderTspecObj) );
		
		pm.Rsvp_Hop_Object.Next_Hop_Address = OI;
		pm.Rsvp_Hop_Object.Logical_Interface_Handle = psbEle->IncInterface;
	}
	
	
	
	void
	ResvMsgPro(ResvMessage *rmess)
	{
		/**
		*** These variables are used to control immediate reservation refreshes.
		**/
		IPAddress Refresh_PHOP_list[InLIST_SIZE];
		int Resv_Refresh_Needed_Flag = OFF,
			NeworMod = OFF;
		IPAddress OI;
		PathStateBlock * psbEle;
		ResvStateBlock * rsbEle;
		ResvStateBlock * activeRSB;
		FlowDescriptorList *fdlist;
		int inx;
		int i;
		int fdinx;
		int newFlag = OFF;
		int psbFoundFlag = OFF;
		GenList locList;
	
		ListInit(&locList);
		/**
		*** Determine the outgoing interface
		**/
		OI = rmess->Rsvp_Hop_Object.Logical_Interface_Handle;
	
		/**
		*** Check the path state
		**/
		psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 1 );
		psbFoundFlag = OFF;
		while( psbEle )
		{
			for(inx=0;inx<InLIST_SIZE;++inx)
			{
					fdlist = &rmess->flow_descriptor_list[inx];
				if( CmpAddress( &psbEle->Sender_Template_Object.SrcAddress,
								&fdlist->Filter_Spec_Object.SrcAddress) &&
					psbEle->Sender_Template_Object.SrcPort !=
					fdlist->Filter_Spec_Object.SrcPort && (
					fdlist->Filter_Spec_Object.SrcPort == 0 ||
					psbEle->Sender_Template_Object.SrcPort == 0 ) )
				{
					/**
					*** Build and send an "Ambiguous Path" PathErr message
					**/
					return;
				}
			}
			if( CmpSession( &rmess->Session_Object, &psbEle->Session_Object ))
				psbFoundFlag = ON;
			psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 0 );
		}
		if( psbFoundFlag == OFF )
		{
			/**
			*** Build and send a ResvErr message specifying "No Path Infomation"
			*** drop the Resv message.
			**/
			return;
		}
		/**
		*** Check for incompatible styles.
		**/
		/**
		*** Process the flow descriptor list to make reservations.
		*** 
		*** Check the path state
		**/
	for(fdinx=0;fdinx<InLIST_SIZE;++fdinx)
	{
		fdlist = &rmess->flow_descriptor_list[fdinx];
		psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 1 );
		while( psbEle )
		{
			if( CmpAddress( &psbEle->Sender_Template_Object.SrcAddress,
							&fdlist->Filter_Spec_Object.SrcAddress) &&
				psbEle->Sender_Template_Object.SrcPort ==
				fdlist->Filter_Spec_Object.SrcPort )
			{
				for(i=0; i<InLIST_SIZE; ++i)  {
					if( CmpAddress( &psbEle->OutInterface_List[i], 
									&OI ) )
					{
						AddListEle( &locList, (char*)psbEle, sizeof(PathStateBlock) );
					}
				}
			}
			psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 0 );
		}
		if( IsListEmpty(&locList) )
		{
			/**
			*** Build and send an error message specifying "No sender information
			**/
			return;
		}
		inx=0;
		psbEle = (PathStateBlock *)GetNextListEle( &locList, 1 );
		while( psbEle )
		{
		int flag;
			flag=OFF;
			for(i=0; i<InLIST_SIZE; ++i)  {
				if( CmpAddress( &Refresh_PHOP_list[i], 
								&psbEle->Previous_Hop_Address) )
				{
					flag=ON;
				}
			}
			if( flag == OFF )
				Refresh_PHOP_list[inx++] = psbEle->Previous_Hop_Address;
			psbEle = (PathStateBlock *)GetNextListEle( &locList, 0 );
		}
		rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 1 );
		activeRSB = NULL;
		while( rsbEle )
		{
			if( CmpSession( &rsbEle->Session_Object, &rmess->Session_Object) &&
				CmpAddress( &rsbEle->Next_Hop_Address,
							&rmess->Rsvp_Hop_Object.Next_Hop_Address) )
			{
				activeRSB = rsbEle;
				break;
			}
			rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 0 );
		}
		if( activeRSB == NULL )
		{
			/** Create a new RSB **/
			activeRSB = (ResvStateBlock *) calloc( 1, sizeof(ResvStateBlock) );
			memcpy( &activeRSB->Session_Object, &rmess->Session_Object,
					sizeof(SessionObj));
			activeRSB->Next_Hop_Address = rmess->Rsvp_Hop_Object.Next_Hop_Address;
			activeRSB->OI = OI;
			memcpy( &activeRSB->Style_Object, &rmess->Style_Object,
					sizeof(StyleObj));
			memcpy( &activeRSB->Filter_Spec_Object, 
					&fdlist->Filter_Spec_Object,
					sizeof(FilterSpecObj));
			memcpy( &activeRSB->Flowspec_Object, 
					&fdlist->Flowspec_Object,
					sizeof(FlowSpecObj));
			NeworMod = ON;
		}
		else
		{
			if( CmpSenderTspec( &activeRSB->Flowspec_Object,
				&fdlist->Flowspec_Object ) )
			{
				memcpy( &activeRSB->Flowspec_Object, 
						&fdlist->Flowspec_Object,
						sizeof(FlowSpecObj));
				NeworMod = ON;
			}
			if( CmpScope( &activeRSB->Scope_Object, &rmess->Scope_Object ) )
			{
				memcpy( &activeRSB->Scope_Object, &rmess->Scope_Object,
						sizeof(ScopeObj) );
				NeworMod = ON;
			}
		}
		/**
		*** Start or restart the cleanup timer on the active RSB.
		**/
		memcpy( &activeRSB->Resv_Confirm_Object, &rmess->Resv_Confirm_Object,
				sizeof(ResvConfirmObj) );
		if( NeworMod == OFF )
			continue;
		/**
		*** Call the UPDATE TRAFFIC CONTROL
		**/
	}
		if( Resv_Refresh_Needed_Flag == ON )
		{
			/**
			*** Send RESV REFRESH for each PHOP in Refresh_PHOP_list.
			**/
		}
		return;
	}
	
	void
	UpdateTrafficControl(ResvStateBlock *activeRSB)
	{
		int Is_Biggest=OFF;
		int TC_E_Police_flag = OFF;
		int Resv_Refresh_Needed_Flag=OFF;
		int inx, i;
		int found;
		FlowSpecObj Tc_Flowspec;
		SenderTspecObj Path_Te;
	
		ResvStateBlock *rsbEle;
		PathStateBlock *psbEle;
		TrafficControlStateBlock * tcsbEle;
		FilterSpecObj TC_Filter_Spec[InLIST_SIZE];
	
		Tc_Flowspec.req_bandwidth = 0;
		Tc_Flowspec.link_delay = 0;
		memset( TC_Filter_Spec, 0, InLIST_SIZE*sizeof(FilterSpecObj) );
	
		rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 1 );
		inx=0;
		while( rsbEle )
		{
			if( CmpSession( &rsbEle->Session_Object, &activeRSB->Session_Object) &&
				CmpFilterSpec( &rsbEle->Filter_Spec_Object, 
							   &activeRSB->Filter_Spec_Object) &&
				CmpAddress(&rsbEle->OI, &activeRSB->OI) )
			{
				/* Compute the effective kernel flowspec */
				if( rsbEle->Flowspec_Object.req_bandwidth > 
					Tc_Flowspec.req_bandwidth )
				{
					Tc_Flowspec.req_bandwidth = 
						rsbEle->Flowspec_Object.req_bandwidth;
				}
				if( rsbEle->Flowspec_Object.link_delay > Tc_Flowspec.link_delay )
				{
					Tc_Flowspec.link_delay = rsbEle->Flowspec_Object.link_delay;
				}
				found=OFF;
				for( i=0; i<inx; ++i )
				{
					if( CmpFilterSpec( &rsbEle->Filter_Spec_Object, 
									   &TC_Filter_Spec[i] ) )
					{
						found=ON;
						break;
					}
				}
				if( found==OFF )
				{
					TC_Filter_Spec[inx++] = rsbEle->Filter_Spec_Object;
				}
			}
			rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 0 );
		}
		if( rsbEle->Flowspec_Object.req_bandwidth >
			Tc_Flowspec.req_bandwidth && rsbEle->Flowspec_Object.link_delay >
			Tc_Flowspec.link_delay )
		{
			Is_Biggest = ON;
		}
		/* Set TC_B_Police_flag */
	
		Path_Te.req_bandwidth = 0;
		Path_Te.link_delay = 0;
		psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 1 );
		while( psbEle )
		{
			if( CmpFilterSpec( &psbEle->Sender_Template_Object,
							   &activeRSB->Filter_Spec_Object ) )
			{
				for(i=0; i<InLIST_SIZE; ++i)  {
					if( CmpAddress( &psbEle->OutInterface_List[i], 
									&activeRSB->OI ) )
					{
						if( psbEle->E_Police_Flag == ON )
						{
							TC_E_Police_flag = ON;
						}
						Path_Te.req_bandwidth +=
							psbEle->Sender_Tspec_Object.req_bandwidth;
						Path_Te.link_delay +=
							psbEle->Sender_Tspec_Object.link_delay;
					}
				}
			}
			psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 0 );
		}
	
		found=OFF;
		tcsbEle = (TrafficControlStateBlock*) GetNextListEle( &TCSBList, 1 );
		while( tcsbEle )
		{
			if( CmpSession( &tcsbEle->Session_Object,
							&activeRSB->Session_Object ) &&
				CmpAddress( &tcsbEle->OI, &activeRSB->OI ) )
			{
				for(i=0; i<InLIST_SIZE; ++i)  {
					if( CmpFilterSpec( &tcsbEle->Filter_Spec_Object[i],
									   &activeRSB->Filter_Spec_Object ) )
					{
						found=ON;
						break;
					}
				}
			}
						
			tcsbEle = (TrafficControlStateBlock*) GetNextListEle( &TCSBList, 0 );
		}
		if( found==OFF )
		{
			/* TCSB is new */
			TrafficControlStateBlock * newTcsb;
			tcsbEle = (TrafficControlStateBlock *) 
					  calloc(1, sizeof( TrafficControlStateBlock ) );
			memcpy( &tcsbEle->Session_Object,
					&activeRSB->Session_Object, 
					sizeof( SessionObj ) );
			CpyAddress( &tcsbEle->OI, &activeRSB->OI );
			memcpy( &tcsbEle->TC_Flowspec,
					&Tc_Flowspec,
					sizeof( FlowSpecObj ) );
			memcpy( tcsbEle->Filter_Spec_Object,
					TC_Filter_Spec,
					InLIST_SIZE*sizeof(FilterSpecObj) );
			memcpy( &tcsbEle->TC_Tspec,
					&Path_Te,
					sizeof( SenderTspecObj ) );
			tcsbEle->E_Police_Flag = TC_E_Police_flag;
		    Resv_Refresh_Needed_Flag=ON;
			/* Make call TC_AddFlowSpec */
			/* If this call fails send ResvErr message specifying 
			   "Admission Control Failed" */	
			/* Otherwise */
			/* Record Rhandle and Fwd_Flowspec in the TCSB. For each 
			   filter_spec F in TC_Filter_Spec * call TC_AddFilter 
			   and record the returned Fhandle in TCSB */
		}
		else
		{
			/* If TCSB is not new */
			if( CmpFlowSpec( &tcsbEle->TC_Flowspec, &Tc_Flowspec ) ||
				CmpSenderTspec( &tcsbEle->TC_Tspec, &Path_Te ) ||
				tcsbEle->E_Police_Flag != TC_E_Police_flag )
			{
				if( CmpFlowSpec( &tcsbEle->TC_Flowspec, &Tc_Flowspec ) ||
					CmpSenderTspec( &tcsbEle->TC_Tspec, &Path_Te ) )
				{
					Resv_Refresh_Needed_Flag=ON;
				}
				/* Modify the reservations */
				/* TC_ModFlowspec */
			}
			if( memcmp( tcsbEle->Filter_Spec_Object,
					TC_Filter_Spec,
					InLIST_SIZE*sizeof(FilterSpecObj) ) )
			{
				/* Make an appropriate set of TC_DelFilter and TC_AddFilter
				   calls to transform the Filter_spec_list in the TCSB
				   into the new TC_Filter_Spec* */
			}
		}
		if ( Is_Biggest == ON )
		{
			tcsbEle->Resv_Confirm_Object = activeRSB->Resv_Confirm_Object;
			Resv_Refresh_Needed_Flag=ON;
		}
		else
		{
			/* Create and send a ResvConf message to the address */
		}
		return;
	}
	
	void
	ResvRefresh( ResvStateBlock *rsbEle )
	{
		int Need_Scope=OFF;
		int B_Merge=OFF;
		int i,j;
		int found;
		IPAddress PH;
	
		ResvMessage outRM;
		FlowSpecObj Tc_Flowspec;
		PathStateBlock *psbEle;
		TrafficControlStateBlock * tcsbEle;
	
		outRM.Session_Object = rsbEle->Session_Object;
		outRM.Time_Values_Object.Refresh_Period = REFRESH_PERIOD;
		outRM.Rsvp_Hop_Object.Next_Hop_Address = rsbEle->Next_Hop_Address;
		outRM.Rsvp_Hop_Object.Logical_Interface_Handle = rsbEle->OI;
		psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 1 );
		while( psbEle )
		{
			PH = psbEle->Previous_Hop_Address;
			Tc_Flowspec.req_bandwidth = 0;
			Tc_Flowspec.link_delay = 0;
			tcsbEle = (TrafficControlStateBlock *)GetNextListEle( &TCSBList, 1 );
			while( tcsbEle )
			{
				found=OFF;
				for( i=0; i<InLIST_SIZE; ++i )
				{
					if( CmpSenderTemplate( &psbEle->Sender_Template_Object,
							&tcsbEle->Filter_Spec_Object[i] ) )
					{
						found=ON;
						break;
					}
				}
				if( found==ON )
				{
					found=OFF;
					for( i=0; i<InLIST_SIZE; ++i )
					{
						if( CmpAddress( &psbEle->OutInterface_List[i],
										&tcsbEle->OI ) )
						{
							found=ON;
							break;
						}
					}
				}
				if( found==ON )
				{
					if( CmpLocalAddress(
						&psbEle->Sender_Template_Object.SrcAddress ) )
					{
						outRM.Resv_Confirm_Object = tcsbEle->Resv_Confirm_Object;
						/* Send the message */
						break;
					}
					if( B_Merge==OFF )
					{
						/* Ignore a blockaded TCSB */
						if ( tcsbEle->Fwd_Flowspec.req_bandwidth >
							 Tc_Flowspec.req_bandwidth )
						{
							Tc_Flowspec.req_bandwidth =
								tcsbEle->Fwd_Flowspec.req_bandwidth;
						}
						if ( tcsbEle->Fwd_Flowspec.link_delay >
							 Tc_Flowspec.link_delay )
						{
							Tc_Flowspec.link_delay =
								tcsbEle->Fwd_Flowspec.link_delay;
						}
					}
				}
				tcsbEle = (TrafficControlStateBlock*)GetNextListEle( &TCSBList, 0 );
			}
			psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 0 );
		}
	}
	
	void
	PathTearPro(PathTearMessage *pmess)
	{
		int found;
		int i;
		PathStateBlock * psbEle = NULL;
		ResvStateBlock * rsbEle = NULL;
	
		found = OFF;
		psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 1 );
		while( psbEle )
		{
			if( CmpSession( &psbEle->Session_Object,
							&pmess->Session_Object ) &&
				CmpSenderTemplate( &psbEle->Sender_Template_Object,
			 			    &pmess->Sender_Descriptor.Sender_Template_Object ) )
			{
				found = ON;
				break;
			}
			psbEle = (PathStateBlock *)GetNextListEle( &PSBList, 0 );
		}
		if( psbEle == NULL )
		{
			return;
		}
		if( found == ON )
		{
			rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 1 );
			while( rsbEle )
			{
				if( CmpFilterSpec( &rsbEle->Filter_Spec_Object,
						(FilterSpecObj *)&psbEle->Sender_Template_Object ))
				{
					for(i=0; i<InLIST_SIZE; ++i)  
					{
						if( CmpAddress( &rsbEle->OI,&psbEle->OutInterface_List[i]) )
						{
							DeleteListEle(&RSBList, (char*)rsbEle, sizeof(ResvStateBlock));
							/* Execute the event sequence UPDATE TRAFFIC CONTROL */
						}
					}
				}
				rsbEle = (ResvStateBlock *)GetNextListEle( &RSBList, 0 );
			}
		}
		DeleteListEle(&PSBList, (char*)psbEle, sizeof(PathStateBlock));
		return;
	}
	
	
	
	void *
	AddListEle( GenList *lst, char *Ele, int size )
	{
		GenList * new;
		GenList * next;
		GenList * prev;
		char * tele;
		if(lst == NULL)
			return NULL;
		if(Ele == NULL)
			return NULL;
		new = (GenList *)calloc(1, sizeof(GenList));
		if(new == NULL)
			PrintErr("No Memory");
		tele = (char *)calloc(1, size);
		if(tele == NULL)
			PrintErr("No Memory");
	    memcpy( tele, Ele, size );
		next = lst;
		prev = lst;
		while(next)
		{
			prev=next;
			next=next->next;
		}
		prev->next = new;
		new->next = NULL;
		new->ele = tele;
		return new;
	}
	
	void
	DeleteListEle( GenList *lst, char *Ele, int size )
	{
		GenList *next;
		GenList *prev;
		if(lst == NULL)
			return ;
		if(Ele == NULL)
			return ;
		prev = lst;
		next = lst;
		while(next)
		{
			if( next->ele != NULL )
			{
				if( !memcmp( Ele, next->ele, size ) )
				{
					prev->next = next->next;
					free(next->ele);
					free(next);
					return;
				}
			}
			prev=next;
			next=next->next;
		}
	}
	
	char *
	GetNextListEle( GenList *lst, int flag )
	{
		GenList * tmp;
		char *ele;
		if(lst == NULL)
			return NULL;
		if( flag )
			lst->curr = lst->next;
		if( lst->curr == NULL )
		{
			ele = NULL;
		}
		else 
		{
			tmp = lst->curr;
			lst->curr = lst->curr->next;
			ele = tmp->ele;
		}
		
		return ele;
	}
	
	void
	ListInit(GenList* lst)
	{
		lst->next=NULL;
		lst->curr=NULL;
		lst->ele=NULL;
	}
	
	int
	IsListEmpty(GenList* lst)
	{
		if(lst->next == NULL)
			return 1;
		return 0;
	}
	
